<?php
/*
 * Plugin: Teamspeak3-Info
 * ~~~~~~~~~~~~~~~~~~~
 *  For a detailed description and documentation, please refer to: https://www.uaseco.org/plugins/repository/undef.de/teamspeak3-info.php
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE: This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------------------------------------------------------
 *
 */

	// Start the plugin
	$_PLUGIN = new PluginTeamspeak3Info();

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

class PluginTeamspeak3Info extends Plugin {
	public $config = array();

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function __construct () {

		// Describe the Plugin
		$this->setVersion('0.9.9');
		$this->setBuild('2018-06-03');
		$this->setAuthor('undef.de');
		$this->setContributors(array('SilentStorm'));
		$this->setCopyright('2009 - 2018 by undef.de');
		$this->setDescription(new Message('plugin.teamspeak3_info', 'plugin_description'));

		// Register events to interact on
		$this->registerEvent('onSync',				'onSync');
		$this->registerEvent('onPlayerConnect',			'onPlayerConnect');
		$this->registerEvent('onPlayerManialinkPageAnswer',	'onPlayerManialinkPageAnswer');
		$this->registerEvent('onEverySecond',			'onEverySecond');
		$this->registerEvent('onBeginMap',			'onBeginMap');
		$this->registerEvent('onEndMap',			'onEndMap');

		$this->registerChatCommand('ts3',			'chat_ts3',	new Message('plugin.teamspeak3_info', 'chat_ts3'),	Player::MASTERADMINS);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function onSync ($aseco) {

		// Read Configuration
		if (!$config = $aseco->parser->xmlToArray('config/teamspeak3_info.xml', true, true)) {
			trigger_error('[Teamspeak3Info] Could not read/parse config file "config/teamspeak3_info.xml"!', E_USER_ERROR);
		}
		$this->config = $config['SETTINGS'];
		unset($config);

		// Check for required <public_host>, <public_port> and <query_port>
		if (!isset($this->config['TS3_SERVER'][0]['PUBLIC_HOST'][0]) || $this->config['TS3_SERVER'][0]['PUBLIC_HOST'][0] === '') {
			trigger_error('[Teamspeak3Info] Missing required option <public_host> in config file "config/teamspeak3_info.xml"!', E_USER_ERROR);
		}
		if (!isset($this->config['TS3_SERVER'][0]['PUBLIC_PORT'][0]) || $this->config['TS3_SERVER'][0]['PUBLIC_PORT'][0] === '') {
			trigger_error('[Teamspeak3Info] Missing required option <public_port> in config file "config/teamspeak3_info.xml"!', E_USER_ERROR);
		}
		if (!isset($this->config['TS3_SERVER'][0]['QUERY_PORT'][0]) || $this->config['TS3_SERVER'][0]['QUERY_PORT'][0] === '') {
			trigger_error('[Teamspeak3Info] Missing required option <query_port> in config file "config/teamspeak3_info.xml"!', E_USER_ERROR);
		}
		if (!isset($this->config['TS3_SERVER'][0]['ID'][0]) || $this->config['TS3_SERVER'][0]['ID'][0] === 0) {
			$this->config['TS3_SERVER'][0]['ID'][0] = 1;			// Set default
		}

		// Setup Query-Host to Public-Host if unset
		if (!isset($this->config['TS3_SERVER'][0]['QUERY_HOST'][0]) || $this->config['TS3_SERVER'][0]['QUERY_HOST'][0] === '') {
			$this->config['TS3_SERVER'][0]['QUERY_HOST'][0] = $this->config['TS3_SERVER'][0]['PUBLIC_HOST'][0];
		}

		$this->config['Timeout']		= 5;
		$this->config['DecodeUtf8']		= false;
		$this->config['Ts3Data']		= array();
		$this->config['RefreshTimestamp']	= time();		// Refresh now
		$this->config['DEBUG']			= ((strtoupper((string)$this->config['DEBUG'][0]) === 'TRUE') ? true : false);
		unset($this->config['DEBUG'][0]);

		$this->config['TS3_SERVER'][0]['HIDE_CHANNELS'][0] = str_replace(' ', '', $this->config['TS3_SERVER'][0]['HIDE_CHANNELS'][0]);
		$this->config['HideChannels']		= explode(',', $this->config['TS3_SERVER'][0]['HIDE_CHANNELS'][0]);
		unset($this->config['DEBUG'][0]);

		// Setup TS3 Escape Sequence
		$this->config['EscapeTable']['ts3']	= array('\\\\', 	"\/", 		"\s", 		"\p", 		"\a", 		"\b", 		"\f", 		"\n", 		"\r", 		"\t", 		"\v");
		$this->config['EscapeTable']['char']	= array(chr(92),	chr(47),	chr(32),	chr(124),	chr(7),		chr(8),		chr(12),	chr(10),	chr(3),		chr(9),		chr(11));
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function onEverySecond ($aseco) {

		// Is it time for refresh the InfoWidgets?
		if (time() >= $this->config['RefreshTimestamp'] && $aseco->server->gamestate === Server::RACE) {
			// Set next refresh timestamp
			$this->config['RefreshTimestamp'] = time() + $this->config['WIDGET'][0]['REFRESH_INTERVAL'][0];

			// Update ServerData
			$this->queryServer();

			// Update InfoWidget at all Players at Race
			$this->buildInfoWidget(false);
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function onPlayerConnect ($aseco, $player) {

		// Send InfoWidget to this Player
		$this->buildInfoWidget($player->login);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function onPlayerManialinkPageAnswer ($aseco, $login, $params) {

		// Get Player
		if (!$player = $aseco->server->players->getPlayerByLogin($login)) {
			return;
		}

		if ($params['Action'] === 'showInfoWindow') {
			// Build InfoWindow
			$widgets = $this->buildInfoWindow($player);

			// Send to given Player
			$aseco->sendManialink($widgets, $player->login);
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function onBeginMap ($aseco, $race) {

		// Show InfoWidget at all Players at Race
		$this->buildInfoWidget(false);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function onEndMap ($aseco, $race) {

		// Hide InfoWidget and InfoWindow at all Players at Score
		$xml = '<manialink id="Teamspeak3InfoWidget" name="Teamspeak3InfoWidget" version="3"></manialink>';

		// Send to all connected Players
		$aseco->sendManialink($xml);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function chat_ts3 ($aseco, $login, $chat_command, $chat_parameter) {

		if (!$player = $aseco->server->players->getPlayerByLogin($login)) {
			return;
		}

		// Init
		$message = false;

		if (strtoupper($chat_parameter) === 'RELOAD') {
			if ($aseco->isMasterAdmin($player)) {
				$aseco->console('[Teamspeak3Info] MasterAdmin ['. $player->login .'] reloads the "config/teamspeak3_info.xml".');
				$message = '{#admin} Reload of the configuration "config/teamspeak3_info.xml" done.';
				$this->onSync($aseco);
			}
		}

		// Show message
		if ($message !== false) {
			$aseco->sendChatMessage($message, $player->login);
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function buildInfoWidget ($login = false) {
		global $aseco;

		$xml = '<manialink id="Teamspeak3InfoWidget" name="Teamspeak3InfoWidget" version="3">';
		$current_online = (isset($this->config['Ts3Data']['server']['virtualserver_clientsonline']) ? $this->config['Ts3Data']['server']['virtualserver_clientsonline'] : '---');
		$current_max = (isset($this->config['Ts3Data']['server']['virtualserver_maxclients']) ? $this->config['Ts3Data']['server']['virtualserver_maxclients'] : '---');

		$xml .= '<frame pos="'. $this->config['WIDGET'][0]['POS_X'][0] .' '. $this->config['WIDGET'][0]['POS_Y'][0] .'" z-index="0" id="Frame_Teamspeak3infoWidget">';
		if ($this->config['WIDGET'][0]['BACKGROUND_DEFAULT'][0] !== '') {
			$xml .= '<quad pos="0 0" z-index="0.001" size="11.5 12.19" bgcolor="'. $this->config['WIDGET'][0]['BACKGROUND_DEFAULT'][0] .'" bgcolorfocus="'. $this->config['WIDGET'][0]['BACKGROUND_FOCUS'][0] .'" scriptevents="1" id="ButtonTeamspeak3InfoWidget"/>';
		}
		else {
			$xml .= '<quad pos="0 0" z-index="0.001" size="11.5 12.19" style="'. $this->config['WIDGET'][0]['BACKGROUND_STYLE'][0] .'" substyle="'. $this->config['WIDGET'][0]['BACKGROUND_SUBSTYLE'][0] .'" scriptevents="1" id="ButtonTeamspeak3InfoWidget"/>';
		}
		$xml .= '<quad pos="-0.45 -8.625" z-index="0.002" size="5.25 3.9375" image="'. $this->config['IMAGES'][0]['WIDGET_OPEN_SMALL'][0] .'"/>';
		$xml .= '<quad pos="2.8125 -0.75" z-index="0.002" size="5.625 5.625" image="'. $this->config['IMAGES'][0]['TEAMSPEAK_LOGO'][0] .'"/>';
		$xml .= '<label pos="5.75 -6.775" z-index="0.1" size="10.6 2.5" halign="center" textsize="1" scale="0.95" text="'. $current_online .'/'. $current_max .'"/>';
		$xml .= '<label pos="5.75 -9.3875" z-index="0.1" size="15.875 3.75" halign="center" textcolor="'. $this->config['WIDGET'][0]['TEXT_COLOR'][0] .'" textsize="1" scale="0.6" text="TEAMSPEAK"/>';
		$xml .= '</frame>';

		$connected = 'True';
		if (!isset($this->config['Ts3Data']['server'])) {
			$connected = 'False';
		}

$xml .= <<<EOL
<script><!--
 /*
 * ==================================
 * Function:	plugin.teamspeak3_info.php
 * Author:	undef.de
 * License:	GPLv3
 * ==================================
 */
main () {
	declare CMlFrame Container	<=> (Page.GetFirstChild("Frame_Teamspeak3infoWidget") as CMlFrame);
	Container.RelativeScale		= {$this->config['WIDGET'][0]['SCALE'][0]};
	declare Boolean Connected	= {$connected};
	while (True) {
		yield;
		if (!PageIsVisible || InputPlayer == Null) {
			continue;
		}

		foreach (Event in PendingEvents) {
			switch (Event.Type) {
				case CMlEvent::Type::MouseOver : {
					Audio.PlaySoundEvent(CAudioManager::ELibSound::Valid, 2, 1.0);
				}
				case CMlEvent::Type::MouseClick : {
					TriggerPageAction("PluginTeamspeak3Info?Action=showInfoWindow");
					if (Connected == False) {
						Audio.PlaySoundEvent(CAudioManager::ELibSound::Alert, 1, 1.0);
					}
					else {
						Audio.PlaySoundEvent(CAudioManager::ELibSound::Valid, 0, 1.0);
					}
				}
			}
		}
	}
}
--></script>
EOL;
		$xml .= '</manialink>';

		if ($login !== false) {
			// Send to given Player
			$aseco->sendManialink($xml, $login);
		}
		else {
			// Send to all connected Players
			$aseco->sendManialink($xml);
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function buildInfoWindow ($player) {
		global $aseco;

		// Bail out if Data not already retrieved
		if (!isset($this->config['Ts3Data']['server'])) {
			return;
		}

		$xml = '<stylesheet>';
		$xml .= '<style class="labels" textsize="1" scale="0.85" textcolor="'. $this->config['WINDOW'][0]['COLORS'][0]['DEFAULT'][0] .'"/>';
		$xml .= '</stylesheet>';

		// BEGIN: Server Overview Box
		$xml .= '<frame pos="0 0" z-index="0">';
		$xml .= '<quad pos="0 0" z-index="0.02" size="48.5 40" bgcolor="FFFFFF33"/>';
		$xml .= '<quad pos="0.5 -0.5" z-index="0.04" size="47.5 4" bgcolor="0099FFFF"/>';
		$xml .= '<quad pos="1.5 -0.5" z-index="0.05" size="3.75 3.75" style="Icons128x128_1" substyle="Statistics"/>';
		$xml .= '<label pos="6 -1.22" z-index="0.05" size="43.25 0" class="labels" text="Server Overview"/>';

		$xml .= '<label pos="2 -6.25" z-index="0.05" size="12.5 3.375" class="labels" text="Status:"/>';
		$xml .= '<label pos="2 -10" z-index="0.05" size="12.5 3.375" class="labels" text="Ping:"/>';
		$xml .= '<label pos="2 -13.75" z-index="0.05" size="12.5 3.375" class="labels" text="Clients:"/>';
		$xml .= '<label pos="2 -17.5" z-index="0.05" size="12.5 3.375" class="labels" text="Channels:"/>';
		$xml .= '<label pos="2 -21.25" z-index="0.05" size="12.5 3.375" class="labels" text="Up since:"/>';
		$xml .= '<label pos="2 -25" z-index="0.05" size="12.5 3.375" class="labels" text="Plattform:"/>';
		$xml .= '<label pos="2 -28.75" z-index="0.05" size="12.5 3.375" class="labels" text="Version:"/>';

		$xml .= '<label pos="14.75 -6.25" z-index="0.05" size="27.5 3.375" class="labels" text=" '. $this->config['Ts3Data']['server']['virtualserver_status'] .'"/>';
		$xml .= '<label pos="14.75 -10" z-index="0.05" size="27.5 3.375" class="labels" text=" '. $this->config['Ts3Data']['server']['virtualserver_total_ping'] .'"/>';
		$xml .= '<label pos="14.75 -13.75" z-index="0.05" size="27.5 3.375" class="labels" text=" '. $this->config['Ts3Data']['server']['virtualserver_clientsonline'] .' / '. $this->config['Ts3Data']['server']['virtualserver_maxclients'] .' / $FD0'. $this->config['Ts3Data']['server']['virtualserver_reserved_slots'] .'"/>';
		$xml .= '<label pos="14.75 -17.5" z-index="0.05" size="27.5 3.375" class="labels" text=" '. $this->config['Ts3Data']['server']['virtualserver_channelsonline'] .'"/>';
		$xml .= '<label pos="14.75 -21.25" z-index="0.05" size="27.5 3.375" class="labels" text=" '. $this->config['Ts3Data']['server']['virtualserver_uptime'] .'"/>';
		$xml .= '<label pos="14.75 -25" z-index="0.05" size="27.5 3.375" class="labels" text=" '. $this->config['Ts3Data']['server']['virtualserver_platform'] .'"/>';
		$xml .= '<label pos="14.75 -28.75" z-index="0.05" size="27.5 3.375" class="labels" text=" '. $this->config['Ts3Data']['server']['virtualserver_version'] .'"/>';
		$xml .= '</frame>';
		// END: Server Overview Box


		// BEGIN: Legend Box
		$xml .= '<frame pos="0 -42" z-index="0.05">';
		$xml .= '<quad pos="0 0" z-index="0.02" size="48.5 23" bgcolor="FFFFFF33"/>';
		$xml .= '<quad pos="0.5 -0.5" z-index="0.04" size="47.5 4" bgcolor="0099FFFF"/>';
		$xml .= '<quad pos="1.5 -0.5" z-index="0.05" size="3.75 3.75" style="Icons128x32_1" substyle="Settings"/>';
		$xml .= '<label pos="6 -1.22" z-index="0.05" size="43.25 0" class="labels" text="Legend"/>';

		$xml .= '<quad pos="2 -6.4375" z-index="0.05" size="3.2 3.2" style="Icons128x128_1" substyle="Lan"/>';
		$xml .= '<quad pos="2 -9.4375" z-index="0.05" size="3.2 3.2" style="Icons64x64_1" substyle="StatePrivate"/>';
		$xml .= '<quad pos="2 -12.4375" z-index="0.05" size="3.2 3.2" style="BgRaceScore2" substyle="Fame"/>';
		$xml .= '<quad pos="2 -15.4375" z-index="0.05" size="3.2 3.2" style="Icons64x64_1" substyle="Buddy"/>';
		$xml .= '<quad pos="2 -18.4375" z-index="0.05" size="3.2 3.2" style="Icons64x64_1" substyle="NotBuddy"/>';
		$xml .= '<label pos="7.5 -6.625" z-index="0.05" size="38.25 3.2" class="labels" text=" Channel (open)"/>';
		$xml .= '<label pos="7.5 -9.625" z-index="0.05" size="38.25 3.2" class="labels" text=" Channel (non-public)"/>';
		$xml .= '<label pos="7.5 -12.625" z-index="0.05" size="38.25 3.2" class="labels" text=" Channel (default)"/>';
		$xml .= '<label pos="7.5 -15.625" z-index="0.05" size="38.25 3.2" class="labels" text=" User"/>';
		$xml .= '<label pos="7.5 -18.625" z-index="0.05" size="38.25 3.2" class="labels" text=" User (AFK)"/>';
		$xml .= '</frame>';
		// END: Legend Box


		// BEGIN: Download Box
		$xml .= '<frame pos="0 -67" z-index="0.05">';
		$xml .= '<quad pos="0 0" z-index="0.02" size="48.5 23" bgcolor="FFFFFF33"/>';
		$xml .= '<quad pos="0.5 -0.5" z-index="0.04" size="47.5 4" bgcolor="0099FFFF"/>';
		$xml .= '<quad pos="1.5 -0.5" z-index="0.05" size="3.75 3.75" style="Icons64x64_1" substyle="TrackInfo"/>';
		$xml .= '<label pos="6 -1.22" z-index="0.05" size="43.25 0" class="labels" text="Please note"/>';

		$xml .= '<quad pos="10 -5.875" z-index="0.04" size="9.375 9.375" halign="center" style="Icons64x64_1" substyle="YellowHigh"/>';
		$xml .= '<label pos="10 -9" z-index="0.05" size="23 0" class="labels" halign="center" textsize="3.5" text="$O$000!"/>';
		$xml .= '<label pos="18 -7.2" z-index="0.05" size="28.75 3.1875" class="labels" autonewline="1" text="It is required, that you'. LF .'have the Teamspeak3'. LF .'client installed!"/>';
		$xml .= '<label pos="24 -18.4375" z-index="0.05" size="44.5 4.3125" class="labels" halign="center" valign="center" url="https://www.teamspeak.com/" focusareacolor1="070F" focusareacolor2="0C0F" text=" "/>';
		$xml .= '<label pos="24 -18.118752" z-index="0.06" size="44.5 4.3125" class="labels" halign="center" valign="center" text="Download Teamspeak3"/>';
		$xml .= '</frame>';
		// END: Download Box


		// BEGIN: Channel and User Box
		$lists = array();
		foreach ($this->config['Ts3Data']['channels'] as &$item) {
			// Hide channels
			if (in_array((int)$item['cid'], $this->config['HideChannels']) === true) {
				continue;
			}
			$lists[$item['cid']]['name']		= $item['channel_name'];
			$lists[$item['cid']]['pid']		= $item['pid'];
			$lists[$item['cid']]['default']		= $item['channel_flag_default'];
			$lists[$item['cid']]['password']	= $item['channel_flag_password'];
		}
		unset($item);
		foreach ($this->config['Ts3Data']['users'] as &$item) {
			// Hide channels
			if (in_array((int)$item['cid'], $this->config['HideChannels']) === true) {
				continue;
			}
			if ((int)$item['client_type'] === 0) {		// reliable detection of any Query Clients ( 0 = regular client, 1 = query client)
				$lists[$item['cid']]['users'][] = array(
					'name'		=> $item['client_nickname'],
					'afk'		=> $item['client_away'],
				);
			}
		}
		unset($item);


		$xml .= '<frame pos="50.5 0" z-index="0.05">';
		$xml .= '<quad pos="0 0" z-index="0.02" size="149.5 90" bgcolor="FFFFFF33"/>';
		$xml .= '<quad pos="0.5 -0.5" z-index="0.04" size="148.5 4" bgcolor="0099FFFF"/>';
		$xml .= '<quad pos="1.5 -0.5" z-index="0.05" size="3.75 3.75" style="Icons128x128_1" substyle="Multiplayer"/>';
		$xml .= '<label pos="6 -1.22" z-index="0.05" size="43.25 0" class="labels" text="Channels and Users"/>';
		$xml .= '</frame>';


		$line = 1;
		$offset = 0;
		$entries = 0;

		$xml .= '<frame pos="50.5 -3" z-index="0">';
		foreach ($lists as &$item) {
			if ($line > 25) {
				$offset += 49.5;
				$line = 1;
			}

			// Indent Subchannels
			$pid_offset = 0;
			if ($item['pid'] > 0) {
				$pid_offset = 2.3;
			}

			$icon = array('style' => 'Icons128x128_1', 'substyle' => 'Lan');	// Open Channel
			if ($item['password'] === 1) {
				// Mark Channel as password protected
				$icon = array('style' => 'Icons64x64_1', 'substyle' => 'StatePrivate');
			}
			else if ($item['default'] === 1) {
				// Mark Channel as default channel
				$icon = array('style' => 'BgRaceScore2', 'substyle' => 'Fame');
			}
			$xml .= '<quad pos="'. ($offset + 43.5) .' -'. (3.28125 * $line) .'" z-index="0.05" size="3.375 3.375" style="'. $icon['style'] .'" substyle="'. $icon['substyle'] .'"/>';
			$xml .= '<label pos="'. ($offset + $pid_offset + 2) .' -'. (3.28125 * $line) .'" z-index="0.05" size="'. (47 - $pid_offset) .' 3.46875" class="labels" textcolor="'. $this->config['WINDOW'][0]['COLORS'][0]['CHANNELS'][0] .'" text="'. htmlspecialchars($aseco->validateUTF8String($item['name'])) .'"/>';

			$line ++;
			$entries ++;

			if (isset($item['users']) && count($item['users']) > 0) {
				foreach ($item['users'] as &$user) {
					$icon = array('style' => 'Icons64x64_1', 'substyle' => 'Buddy');	// Talking User
					if ($user['afk'] === 1) {
						// Mark user as "away from keyboard"
						$icon = array('style' => 'Icons64x64_1', 'substyle' => 'NotBuddy');
					}
					$xml .= '<quad pos="'. ($offset + 43.5) .' -'. (3.28125 * $line) .'" z-index="0.05" size="3.375 3.375" style="'. $icon['style'] .'" substyle="'. $icon['substyle'] .'"/>';
					$xml .= '<label pos="'. ($offset + 6) .' -'. (3.28125 * $line) .'" z-index="0.05" size="42 3.46875" class="labels" textcolor="'. $this->config['WINDOW'][0]['COLORS'][0]['USERS'][0] .'" text="'. htmlspecialchars($aseco->validateUTF8String($user['name'])) .'"/>';

					$line ++;
					$entries ++;

					if ($entries >= 75) {
						break;
					}
				}
				unset($user);
			}
			if ($entries >= 75) {
				break;
			}
		}
		unset($item);
		$xml .= '</frame>';
		// END: Channel and User Box


		// Build Link to the Teamspeak3-Server
		// http://www.teamspeak.com/?page=faq&cat=ts3server#ts3server_weblink
		// ts3server://ts3.hoster.com/?port=9987&nickname=UserNickname&password=serverPassword&channel=MyDefaultChannel&channelpassword=defaultChannelPassword&token=TokenKey&addbookmark=1
		$url = 'ts3server://'. $this->config['TS3_SERVER'][0]['PUBLIC_HOST'][0] .'/?port='. $this->config['TS3_SERVER'][0]['PUBLIC_PORT'][0] .'&amp;nickname='. rawurlencode($aseco->stripStyles($player->nickname, true));
		if ($this->config['TS3_SERVER'][0]['DEFAULT_CHANNEL'][0] !== '') {
			$url .= '&amp;channel='. rawurlencode($this->config['TS3_SERVER'][0]['DEFAULT_CHANNEL'][0]);
		}


		// Setup settings for Window
		$settings_styles = array(
			'icon'			=> 'Icons128x128_1,Buddies',
		);
		$settings_content = array(
			'title'			=> $this->handleSpecialChars($this->config['Ts3Data']['server']['virtualserver_name']) .' - Teamspeak3 Info',
			'data'			=> array($xml),
			'mode'			=> 'pages',
		);
		$settings_footer = array(
			'about_title'		=> 'TEAMSPEAK3-INFO/'. $this->getVersion(),
			'about_link'		=> 'https://www.uaseco.org/plugins/repository/undef.de/teamspeak3-info.php',
			'button_title'		=> 'TALK AT THIS SERVER',
			'button_link'		=> $url,
		);

		$window = new Window();
		$window->setStyles($settings_styles);
		$window->setContent($settings_content);
		$window->setFooter($settings_footer);
		$window->send($player, 0, false);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function handleSpecialChars ($string) {
		global $aseco;

		// Remove links, e.g. "$(L|H|P)[...]...$(L|H|P)"
		$string = preg_replace('/\${1}(L|H|P)\[.*?\](.*?)\$(L|H|P)/i', '$2', $string);
		$string = preg_replace('/\${1}(L|H|P)\[.*?\](.*?)/i', '$2', $string);
		$string = preg_replace('/\${1}(L|H|P)(.*?)/i', '$2', $string);

		// Remove $S (shadow)
		// Remove $H (manialink)
		// Remove $W (wide)
		// Remove $I (italic)
		// Remove $L (link)
		// Remove $O (bold)
		// Remove $N (narrow)
		$string = preg_replace('/\${1}[SHWILON]/i', '', $string);

		// Convert &
		// Convert "
		// Convert '
		// Convert >
		// Convert <
		$string = str_replace(
				array(
					'&',
					'"',
					"'",
					'>',
					'<'
				),
				array(
					'&amp;',
					'&quot;',
					'&apos;',
					'&gt;',
					'&lt;'
				),
				$string
		);
		$string = $aseco->stripNewlines($string);

		return $aseco->validateUTF8String($string);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	// The next functions are based upon tsstatus.php from http://tsstatus.sebastien.me/
	// and was optimized by SilentStorm to our needs
	public function queryServer () {
		global $aseco;

		$socket = @fsockopen($this->config['TS3_SERVER'][0]['QUERY_HOST'][0], $this->config['TS3_SERVER'][0]['QUERY_PORT'][0], $errno, $errstr, $this->config['Timeout']);
		if ($socket !== false) {
			// Checking for a TS3-Server
			$response = $this->sendCommand($socket, 'version');
			if (substr($response, 0, 3) !== 'TS3') {
				$aseco->console('[Teamspeak3Info] Server at "'. $this->config['TS3_SERVER'][0]['QUERY_HOST'][0] .'" is not a Teamspeak3-Server or you have setup a bad query-port!');
				fclose($socket);
				return;
			}

			if (isset($this->config['TS3_SERVER'][0]['QUERY_USER'][0]) && !empty($this->config['TS3_SERVER'][0]['QUERY_USER'][0]) && !is_numeric($this->config['TS3_SERVER'][0]['QUERY_USER'][0]) && $this->config['TS3_SERVER'][0]['QUERY_USER'][0] !== false && isset($this->config['TS3_SERVER'][0]['QUERY_PASS'][0]) && !empty($this->config['TS3_SERVER'][0]['QUERY_PASS'][0]) && !is_numeric($this->config['TS3_SERVER'][0]['QUERY_PASS'][0]) && $this->config['TS3_SERVER'][0]['QUERY_PASS'][0] !== false) {
				$authentication = $this->sendCommand($socket, 'login client_login_name='. $this->escape($this->config['TS3_SERVER'][0]['QUERY_USER'][0]) .' client_login_password='. $this->escape($this->config['TS3_SERVER'][0]['QUERY_PASS'][0]));
				if (stripos($authentication, "error id=0") === false) {
					$aseco->console('[Teamspeak3Info] Failed to authenticate with TS3 Server! Make sure you have set the correct <query_user> and <query_pass> in "config/teamspeak3_info.xml"');
					return;
				}
			}
//			else {
//				$aseco->console('[Teamspeak3Info] Login and/or Password not specified, logging in as "Guest".');
//			}

			$response = '';
			$response .= $this->sendCommand($socket, 'use sid=' . $this->config['TS3_SERVER'][0]['ID'][0]);
			if (!empty($this->config['TS3_SERVER'][0]['QUERY_FRIENDLY_NICKNAME'][0]) && isset($this->config['TS3_SERVER'][0]['QUERY_FRIENDLY_NICKNAME'][0])) {
				$this->sendCommand($socket, 'clientupdate client_nickname=' . $this->escape($this->config['TS3_SERVER'][0]['QUERY_FRIENDLY_NICKNAME'][0]));
			}
			$response .= $this->sendCommand($socket, 'serverinfo');
			$response .= $this->sendCommand($socket, 'channellist -topic -flags -voice -limits');
			$response .= $this->sendCommand($socket, 'clientlist -uid -away -voice -groups');
			if ($socket !== false) {
				@fwrite($socket, "quit\n");
				fclose($socket);
			}

			if ($this->config['DecodeUtf8'] === true) {
				$response = utf8_decode($response);
			}

			if (strpos($response, 'error id=0 msg=ok') !== false) {
				$lines = explode("error id=0 msg=ok\n\r", $response);
				if (count($lines) >= 3) {
					$serverdata = $this->parseLine($lines[1]);
					$this->config['Ts3Data']['server'] = $serverdata[0];
					$this->config['Ts3Data']['channels'] = $this->parseLine($lines[2]);
					$this->config['Ts3Data']['users'] = $this->parseLine($lines[3]);

					// Subtract reserved slots
					$this->config['Ts3Data']['server']['virtualserver_maxclients'] -= $this->config['Ts3Data']['server']['virtualserver_reserved_slots'];

					// Make ping value int
					$this->config['Ts3Data']['server']['virtualserver_total_ping'] = intval($this->config['Ts3Data']['server']['virtualserver_total_ping']);

					// Format the Date of server startup
					$this->config['Ts3Data']['server']['virtualserver_uptime'] = date('Y-m-d H:i:s', (time() - $this->config['Ts3Data']['server']['virtualserver_uptime']) );

					// Always subtract all Query Clients
					$this->config['Ts3Data']['server']['virtualserver_clientsonline'] -= $this->config['Ts3Data']['server']['virtualserver_queryclientsonline'];
				}
				else if ($this->config['DEBUG'] === true) {
					$aseco->console('[Teamspeak3Info] The TS3 Server returned not enough lines: '. LF . $response);
				}
			}
			else if ($this->config['DEBUG'] === true) {
				$aseco->console('[Teamspeak3Info] The TS3 Server returned an error: '. LF . $response);
			}
		}
		else {
			$aseco->console('[Teamspeak3Info] Unable to connect with TS3 Server '. $this->config['TS3_SERVER'][0]['QUERY_HOST'][0] .':'. $this->config['TS3_SERVER'][0]['QUERY_PORT'][0] .' Socket error: '. $errstr .' ['. $errno .']');
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function sendCommand ($socket, $cmd) {
		global $aseco;

		if ($socket === false) {
			return;
		}

		// Send command
		@fwrite($socket, "$cmd\n");

		stream_set_blocking($socket, 0);
		stream_set_timeout($socket, ceil($this->config['Timeout'] / 2));
		$meta = stream_get_meta_data($socket);

		$response = '';
		while (!feof($socket) && !$meta['timed_out'] && strpos($response, 'error id=') === false) {
			$response .= fread($socket, 32);
			$meta = stream_get_meta_data($socket);
		}
		if ($meta['timed_out'] === true) {
			fclose($socket);
			$socket = false;
			$aseco->console('[Teamspeak3Info] TS3 Server timed out, stopping communication for a moment!');
		}
//		if (strpos($response, 'error id=0') === false) {
//			//throw new Exception("TS3 Server returned the following error: " . $this->unescape(trim($response)));
//		}
		return $response;
	}


	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function parseLine ($rawLine) {

		$datas = array();
		$rawItems = explode('|', $rawLine);

		foreach ($rawItems as &$rawItem) {
			$rawDatas = explode(' ', $rawItem);
			$tempDatas = array();
			foreach ($rawDatas as &$rawData) {
				$ar = explode("=", $rawData, 2);
				$tempDatas[$ar[0]] = isset($ar[1]) ? $this->unescape($ar[1]) : '';
			}
			$datas[] = $tempDatas;
		}
		unset($rawItem, $rawData);

		return $datas;
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function escape ($str) {
		return str_replace($this->config['EscapeTable']['char'], $this->config['EscapeTable']['ts3'], $str);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function unescape ($str) {
		return str_replace($this->config['EscapeTable']['ts3'], $this->config['EscapeTable']['char'], $str);
	}
}

?>
